/****************************************************************************
*RevHead_Start                                                              *
*                                                                           *
* Project Details                                                           *
* ---------------                                                           *
*Project Name:    wordmatch                                                 *
*Project Version: 1.0                                                       *
*Copyright Owner: David M. Balean <linprogs@balean.net>                     *
*Copyright Notice:                                                          *
* This program is free software; you can redistribute it and/or modify it   *
* under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License or,         *
* (at your option) any later version.                                       *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
* See the GNU General Public License for more details.                      *
* You should have received a copy of the GNU General Public License         *
* along with this program. If not, see <http://www.gnu.org/licenses/>.      *
*                                                                           *
*Project Date:    Thursday 25th April 2013                                  *
*                                                                           *
*Start DateStamp: 17-Jun-2011 12.37.31 -36000 0                             *
*                                                                           *
*Contact:                                                                   *
* email: linprogs@balean.net                                                *
* snail: D.M. Balean, 44 Wyong Road, Killarney Vale, N.S.W. 2261, AUSTRALIA *
* URL:   http://balean.net/linprogs/index.html                              *
*                                                                           *
*Project Summary:                                                           *
* Match words using wildcards with optional synonyms                        *
* or provide anagrams using wildcards with optional synonyms                *
*                                                                           *
*****************************************************************************
*                                                                           *
*This File:       main.c                                                    *
*Author:          David M. Balean <linprogs@balean.net>                     *
*DateStamp:       25-Apr-2013 12.34.51 -36000 0                             *
*                                                                           *
*File Summary:                                                              *
* Entry point                                                               *
*                                                                           *
*RevHead_End                                                                *
****************************************************************************/

#include <stdlib.h>
#include <sys/types.h>
#include <pwd.h>
#include <ctype.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "about.h"
#include "misc.h"
#include "prefs.h"
#include "match.h"
#include "RH_version.h"


/*** CONSOLE INFO ***********************************
// Attributes
#define RESET       0
#define BRIGHT      1
#define DIM         2
#define UNDERLINE   4
#define BLINK       5
#define REVERSE     7
#define HIDDEN      8
// Fg Colours - NB for Bg Colours add 10
#define BLACK       30
#define RED         31
#define GREEN       32
#define YELLOW      33
#define BLUE        34
#define MAGENTA     35
#define CYAN        36
#define WHITE       37
***/

const char *red_text = "\033[0;1;31m";
const char *normal_colour = "\033[0m";

GtkWidget *main_window;

static GtkWidget *pattern_entry = NULL;
static GtkWidget *anagram_entry = NULL;

static GtkWidget *pattern_label;
static GtkWidget *anagram_label;

static GtkWidget *case_sensitive_check_button = NULL;
static GtkWidget *synonyms_check_button = NULL;
static GtkWidget *match_radio_button = NULL;
static GtkWidget *anagram_radio_button = NULL;
static GtkWidget *with_search_check_button = NULL;
static GtkWidget *include_hyphens_check_button = NULL;

static GtkWidget *close_button;
static GtkWidget *preferences_button;
static GtkWidget *search_button;


static char *action_txt[4] =
  {
  "Search",
  "Preferences",
  "About",
  "Close"
  };

static void callback_search(GtkWidget *widget, gpointer data);
static void callback_quit(GtkWidget *widget, gpointer data);

// Array of pointers to functions
static void (*action[4])(GtkWidget *widget, gpointer data) =
  {
  callback_search,
  callback_preferences,
  callback_about,
  callback_quit
  };

static GtkWidget *menu;
// static GtkWidget *menu_bar;
// static GtkWidget *root_menu;
static GtkWidget *menu_item;

GSList *group = NULL;

GdkPixbuf *rw_pixbuf = NULL;
GdkPixbuf *mw_pixbuf = NULL;
GdkPixbuf *pw_pixbuf = NULL;
GdkPixbuf *ab_pixbuf = NULL;
GtkWidget *status_label = NULL;
char *status_txt = "STATUS: ";
char *wordmatch_prefs_pn_txt = NULL;
                                  
char *wordbuffer = NULL;
char *wordbuffer_end = NULL;

GtkWidget *sw;
GtkWidget *text;
GtkTextBuffer *tbuf;
GList *child_windows = NULL;

char *asterisks_txt = "**********************************************************************\n";

void callback_destroy_child_window(GtkWidget *widget, gpointer data);

void print_colour(char *str, const char *colour)
{
g_print("%s%s%s", colour, str, normal_colour);
}


void free_glist(GList **list)
{
GList *gl = *list;
  
if (gl)
	{
	while (gl)
		{
		g_free(gl->data);
		gl = gl->next;
		}
	g_list_free(*list);
	*list = NULL;
	}
}


/* Clean up word data from result window */
void clean_up(void)
{
/* NOTE: when the program terminates, the child windows self-close */
free_glist(&child_windows);

if (rw_pixbuf)
	{
	g_object_unref(G_OBJECT(rw_pixbuf));
	rw_pixbuf = NULL;
	}	
if (mw_pixbuf)
	{
	g_object_unref(G_OBJECT(mw_pixbuf));
	mw_pixbuf = NULL;
	}
if (pw_pixbuf)
	{
	g_object_unref(G_OBJECT(pw_pixbuf));
	pw_pixbuf = NULL;
	}
if (ab_pixbuf)
	{
	g_object_unref(G_OBJECT(ab_pixbuf));
	ab_pixbuf = NULL;
	}

if (words_pn_txt) g_free(words_pn_txt);
if (synonyms_command_txt) g_free(synonyms_command_txt);
if (wordbuffer) g_free(wordbuffer);
if (wordmatch_prefs_pn_txt) g_free(wordmatch_prefs_pn_txt);
}

static void callback_quit(GtkWidget *widget, gpointer data)
{
clean_up();
gtk_main_quit();
}

// The data here is the result dialog window that has been destroyed
void callback_destroy_child_window(GtkWidget *widget, gpointer data)
{
GList *node = child_windows;
GList *nextnode;

while (node)
	{
	nextnode = node->next;
  
	if (node->data == data)
		{
		g_free(data);
		child_windows = g_list_remove(child_windows, data);
		break;
		}
	node = nextnode;
	}
}


char *get_command(char *command)
{
size_t size = 2048;
char *buffer = NULL;
size_t len = 0;

if ( (buffer = (char *)g_malloc0(size) ) )
	{
	FILE *fp;
	char pipebuffer[PIPE_BUF];
	size_t pipelen;


	if ( (fp = popen(command, "r")) )
		{
	    while ( fgets(pipebuffer, PIPE_BUF, fp) )
	      	{
			pipelen = strlen(pipebuffer);
		
			if ( (pipelen + len) >= size )
		  		{
		    	/* We need to enlarge the buffer! */
		    	size_t newsize = size +(4 * PIPE_BUF);
		    	char *newbuffer = (char *)g_malloc0(newsize);
		  
		    	if (newbuffer)
		      		{
					memcpy(newbuffer, buffer, size);
					g_free(buffer);
					buffer = newbuffer;
					size = newsize;
		      		}
		    	else
		      		{
					/* ERROR: Can't allocate memory */
		      		}		     
		  		}

			strcat(buffer, pipebuffer);
			len += pipelen;
	      	}
	    pclose(fp);
	  	}
	else
	  	{
	    /* ERROR: can't open pipe */
	  	}
  }
else
	{
	/* ERROR: Can't allocate memory */
     }

/* This assumes no zero in the file, i.e. a text file of some kind */
if (buffer)
	{
	char *newbuffer;
	if ( (newbuffer = make_string(buffer)) ) /* remove unwanted memory */
	  	{
	    g_free(buffer);
	    buffer = newbuffer;
	  	}
      }
return(buffer);
}

void create_tags(GtkTextBuffer *buffer)
{
gtk_text_buffer_create_tag (buffer,
			      "red_text",
			      "style",
			      PANGO_STYLE_NORMAL,
			      "foreground",
			      "red",
			      "family",
			      "monospace",
			      NULL);

gtk_text_buffer_create_tag (buffer,
			      "black_text",
			      "style",
			      PANGO_STYLE_NORMAL,
			      "foreground",
			      "black",
			      "family",
			      "monospace",
			      NULL);

gtk_text_buffer_create_tag (buffer, "bold",
                 "weight", PANGO_WEIGHT_BOLD, NULL);  

}

/* AND clean up of any existing data! */
void create_result_window(void)
{	
GtkWidget *result_window, **cw;

GtkWidget *wid;
GtkWidget *hbox;
	
result_window = gtk_dialog_new();
  
gtk_widget_show(result_window);
gtk_window_set_title (GTK_WINDOW (result_window), "WORDS");
gtk_container_set_border_width(GTK_CONTAINER (result_window), 0);
gtk_window_set_modal(GTK_WINDOW(result_window), FALSE);
gtk_window_set_default_size(GTK_WINDOW(result_window), results_window_width, results_window_height); /* w, h */
gtk_window_resize(GTK_WINDOW(result_window),  results_window_width, results_window_height); /* w, h */

if (rw_pixbuf) gtk_window_set_icon( GTK_WINDOW(result_window), rw_pixbuf);

cw = g_malloc(sizeof(GtkWidget *));
*cw = result_window;

child_windows = g_list_append(child_windows, cw);

g_signal_connect(G_OBJECT(result_window), "destroy",
              G_CALLBACK(callback_destroy_child_window),
              (gpointer)cw);


text = gtk_text_view_new();
tbuf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text));
gtk_widget_show (text);
gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
create_tags (tbuf);
sw = gtk_scrolled_window_new(NULL, NULL);

    /* the policy is one of GTK_POLICY AUTOMATIC, or GTK_POLICY_ALWAYS.
     * GTK_POLICY_AUTOMATIC will automatically decide whether you need
     * scrollbars, whereas GTK_POLICY_ALWAYS will always leave the scrollbars
     * there.  The first one is the horizontal scrollbar, the second, 
     * the vertical. */
gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                               GTK_POLICY_AUTOMATIC,
								GTK_POLICY_AUTOMATIC );

gtk_widget_show (sw);
gtk_container_add(GTK_CONTAINER (sw), text);
gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(result_window))), sw, TRUE, TRUE, 0);


hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
gtk_box_set_homogeneous (GTK_BOX(hbox), FALSE);
	
gtk_widget_show(hbox);
gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(result_window))), hbox, FALSE, FALSE, 0);

wid = gtk_label_new(status_txt);
gtk_misc_set_alignment(GTK_MISC (wid), 0, .5);
gtk_widget_show(wid);
gtk_box_pack_start(GTK_BOX(hbox), wid, FALSE, FALSE, 0);

status_label = gtk_label_new("");
gtk_misc_set_alignment(GTK_MISC (status_label), 0, .5);
gtk_widget_show(status_label);
gtk_box_pack_start(GTK_BOX(hbox), status_label, FALSE, FALSE, 0);

}





/*
Place "string" at end of displayed words and append a newline
using colour "type" 0=black (normal), 1=red (dictionary word)
*/
void append_string(char *string, int type)
{
GtkTextIter iter;
	
gtk_text_buffer_get_iter_at_offset(GTK_TEXT_BUFFER(tbuf),
                                   &iter,
                                   -1);		

if (type) // Insert this dictionary entry with a newline
  {
  gtk_text_buffer_insert_with_tags_by_name(tbuf, &iter,
                                           string, -1, "red_text", "bold", NULL);
  gtk_text_buffer_get_iter_at_offset(GTK_TEXT_BUFFER(tbuf),
                                   &iter,
                                   -1);		
  gtk_text_buffer_insert(GTK_TEXT_BUFFER(tbuf),
                       &iter,
                       "\n",
                       -1);	
  }
else
  {
  gtk_text_buffer_insert_with_tags_by_name(tbuf, &iter,
                                           string, -1, "black_text", NULL);
  }

}

void insert_newlines(char *string)
{
char *p = string;  // Current position in string
int n = 0;         // The position in string of p.
int i = 0;         // The position of p in the current line
int c = -1 ;       // Position in string of latest comma


//char *end = &string[strlen(string)];

while (*p)    // As long as there is a current character
	{

  while ((*p)== ' ')
    {
//    if ( (*(p-1)) == '\n') *p = '\r';
    p++;
    i++;
    n++;
    }

  if (*p=='\n')
		{
		p++;
    n++;
    i = 0;
    c = -1;
    continue;
		}

    
  if ( (i >= 70) && (c >= 0) )
    {
    // Use the value in c
    p = &string[c];        // the ',' to be replaced by a newline
    *p = '\n';
    p++;
    n = c + 1;
    i = 0;
    c = -1;
    continue;
    }

  if (*p == ',')
    {
    c = n;                // remember this position in p1
    }

  p++;
  n++;
  i++;
	}

}



// is_anagram to permit any printable character
// EXCEPT  *   [   ]   (   )
// Use if ( isprint(c) ) to check
// then eliminate the exceptions and - if required

// This creates a histogram of the letters in check_letters then removes
// the letters of the check_word. If none remain then it is an anagram.

gboolean is_anagram(char check_letters[], char check_word[], char **error)
{
unsigned int i, len;
char letters[strlen(check_letters)];
char word[strlen(check_word)];
int numbytes = 127 - 32; // All the printable characters

//    = 10 + 26 + ((gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_hyphens_check_button))) ? 0 : 1);

// The histogram - if silent hyphens not selected, add 1 to include hyphens in the histogram
//               - if selected, remove hyphens from letters and word
int table[numbytes];

// Initialize the histogram to zero
memset( table, 0, numbytes*sizeof(int) );


//if (numbytes == 10 + 26) // remove hyphens from letters and word

if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_hyphens_check_button)))
  {
  char *w = word;
  char *t = check_word;

  for (;;)   /* get the word with no hyphens */
    {
    if (*t == '-') t++;
    else
      {
      *w++ = *t;
      if (*t == 0) break;
      else t++;
      }
    }

  w = letters;
  t = check_letters;

  for (;;)   /* get the pattern with no hyphens */
    {
    if (*t == '-') t++;
    else
      {
      *w++ = *t;
      if (*t == 0) break;
      else t++;
      }
    }
  }

else
  {
  strcpy(word, check_word);
  strcpy(letters, check_letters);
  }



len = strlen(letters);

// Read letters, incrementing the histogram entry of that character
for (i = 0; i < len; i++)
  {
  // Must be printable and legal
  // ILLEGAL  ?   *   [   ]   (   )

//==============================================================================
// TRY ALLOWING ? for unknown letter
//==============================================================================
		
  if ( isprint(letters[i]) &&
      (!((letters[i]=='*')||(letters[i]=='[')||(letters[i]==']')||(letters[i]=='(')||(letters[i]==')'))))
    {
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)))
        table[letters[i]-' ']++;
    else
      {
      table[tolower(letters[i])-' ']++;
      }
    }
  else
    {
    // ERROR
    static char anagram_error[128];

    strcpy(anagram_error,  "Anagram error: Illegal character : ");
    if (isprint(letters[i]))
      sprintf(&anagram_error[strlen(anagram_error)], "\"%c\"\n", letters[i]);
    else sprintf(&anagram_error[strlen(anagram_error)], "Unprintable character %02x\n", letters[i]);
    *error = anagram_error;
    return FALSE;
    }
  }




// The histogram of the anagram letters appears OK
{
  static gboolean done = FALSE;
  int n;
  if (!done)
    {
    g_print("HISTOGRAM - numbytes= %d\n", numbytes);

    for (n=0; n<numbytes; n++)
      {
      g_print("%c : %d\n", ' ' + n, table[n]);
      }
    done = TRUE;
    }
}

//exit(0);

len = strlen(word);  // the check_word


// If the histogram contains ? then, if a character in the word isn't found
// in the histogram, it should check for ? in the histogram. I ? is found
// then it should be decremented and then move onto the next letter as if
// it were an ordinary found character.

for (i = 0; i < len; i++)
  {
  // Must be printable and probably should be legal also (could change that if a word found to misbehave)
  // ILLEGAL  ?   *   [   ]   (   )

		
//==============================================================================
// TRY ALLOWING ? for unknown letter
//==============================================================================

  if ( isprint(word[i]) &&
      (! ( (word[i]=='*') || (word[i]=='?') || (word[i]=='[') || (word[i]==']') || (word[i]=='(') || (word[i]==')') ) ) )
    {
	
    if (word[i]=='-')    // Character is '-', special case
      {
      if (!(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_hyphens_check_button))))
        {
        if (table[word[i]-' '] == 0)     // this character isn't in the histogram
					{
					if (table[tolower('?')-' '] == 0)	 // check for '?' in the histogram		
	    			return FALSE;

					else table[tolower('?')-' ']--;     // deduct 1 from the number of '?' characters in the histogram
					}
        else table[word[i]-' ']--;       // deduct 1 from the number of this character in the histogram
        }
      }


		// Character isn't '-' and is case sensitive	
    else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)))
      {
      if (table[word[i]-' '] == 0)       // this character isn't in the histogram
				{
				if (table[tolower('?')-' '] == 0)	 // check for '?' in the histogram		
					return FALSE;
				else table[tolower('?')-' ']--;    // deduct 1 from the number of '?' characters in the histogram
				}
				
			else table[word[i]-' ']--;    // Got it so deduct 1 from the number of this character in the histogram
      }


		// Character isn't '-' and is not case sensitive	
    else if (table[tolower(word[i])-' '] == 0)   // this character isn't in the histogram
			{
			if (table[tolower('?')-' '] == 0)	 // check for '?' in the histogram		
					{
					if (table[tolower('?')-' '] == 0)	 // check for '?' in the histogram		
	    			return FALSE;

					else table[tolower('?')-' ']--;     // deduct 1 from the number of '?' characters in the histogram
					}

  		else
				{
				table[tolower('?')-' ']--;    // deduct 1 from the number of '?' characters in the histogram
				}
			}
		else table[tolower(word[i])-' ']--;    // Got it so deduct 1 from the number of this character in the histogram
				
    }

	
  else return FALSE;     // illegal letter in dictionary!
  } // End of for

for (i = 0; i < numbytes; i++)
  {
  if (table[i]) return FALSE; // check that nothing is left over...
  }

return TRUE;
}












/*
 * Return TRUE if we have the word list
 * otherwise output error and return FALSE
*/
gboolean got_wordbuffer(void)
{
if (wordbuffer) return TRUE;
else
  {
	char buffer[128];
	strcpy(buffer, "WORDLIST is empty!");
	gtk_label_set_text(GTK_LABEL(status_label), buffer);

	g_print("%s%s\n", status_txt, buffer);
  return FALSE;
  }
}



void found_word(char *p)
{
/* FOUND A WORD */

print_colour(p, red_text);  // This should work but may not on some systems 
g_print("\n");
append_string(p, 1);
							
if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(synonyms_check_button)))
  {
	char *synonyms;
	char command[strlen(p) + 64];

	sprintf(command, synonyms_command_txt, p);
	synonyms = get_command(command);
	if (strstr(synonyms, "No synonyms known"))
    {
		append_string("<<- No synonyms known ->>\n", 0);
    g_print("<<- No synonyms known ->>\n");
    }
	else
    {
		insert_newlines(synonyms);
		append_string(synonyms, 0);
    g_print("%s", synonyms);
    }
	g_free(synonyms);
	append_string(asterisks_txt, 0);
  g_print("%s", asterisks_txt);
  }
}

gboolean hyphens_ok(char *word, char *pattern)
{
char *w=word, *p=pattern;
int num_w=0, num_p=0;

while (*w)
  {
  if ( (*w) == '-' ) num_w++;
  w++;
  }
while (*p)
  {
  if ( (*p) == '-' ) num_p++;
  p++;
  }
return (num_w >= num_p);
}



//==============================================================================
//                    SEARCH FOR ANAGRAMS
//==============================================================================





// Go through all words in the dictionary and check if it is an anagram
// This could probably be done quicker...
void search_for_anagrams(void)
{
int num_words = 0;


// Are we using a pattern?	
char *letters = (char *)gtk_entry_get_text(GTK_ENTRY(anagram_entry));
	
char *anagram_error = NULL;
gulong p_error = 0;

g_print("search for anagrams\n");

g_print("%s\n", letters);

// Go through the dictionary
// Omit the check if not the correct length


if (letters && *letters)
  {
	char *p;  /* position in wordbuffer */
	char *t;  /* terminal zero of current word being checked */

  char buffer[strlen(letters) + 256];
  sprintf(buffer, "Anagrams matching letters:-\n           \"%s\"\n", letters);
  g_print("%s", buffer);
  append_string(buffer, 0);


  p = t = wordbuffer;

  while (t < wordbuffer_end)
    {
		if (*t == '\n')
      {
			*t = '\0';

      if (is_anagram(letters, p, &anagram_error))
        {
				/* FOUND A WORD */

// Now if we are comparing with a pattern
        if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (with_search_check_button)))
          {
	        char *pattern = (char *)gtk_entry_get_text(GTK_ENTRY(pattern_entry));

          if (!(pattern && *pattern))
		        {
		        char buffer[128];
		        strcpy(buffer, "Pattern to find is empty!");
		        gtk_label_set_text(GTK_LABEL(status_label), buffer);
		        g_print("%s%s\n", status_txt, buffer);
		        }

	        else
            {
				    if (match(  p,            /* word           */
								pattern,              /* pattern        */
								!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)), /*FALSE = case sensitive */
								NULL,         /* eos            */
								&p_error) )   /* error number   */
              {
					    /* FOUND A WORD */
              found_word(p);
              num_words++;

              }
				    else if (p_error)
              {
			        break;
              }
        

            else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_hyphens_check_button)))
              {
              // Copy the pattern and the word, remove hyphens from both copies then check
              char word_no_hyphen[strlen(p) + 1];
              char pattern_no_hyphen[strlen(pattern) + 1];

              char *w = word_no_hyphen;
              char *t = p;

              for (;;)   /* get the word with no hyphens */
                {
                if (*t == '-') t++;
                else
                  {
                  *w++ = *t;
                  if (*t == 0) break;
                  else t++;
                  }
                }

              w = pattern_no_hyphen;
              t = pattern;

              for (;;)   /* get the pattern with no hyphens */
                {
                if (*t == '-') t++;
                else
                  {
                  *w++ = *t;
                  if (*t == 0) break;
                  else t++;
                  }

                }

              if (match(  word_no_hyphen,       /* word           */
								pattern_no_hyphen,          /* pattern        */
								!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)), /*FALSE = case sensitive */
								NULL,         /* eos            */
								&p_error) )   /* error number   */
                {
					      /* FOUND A WORD */
                // Re-check the hyphens here
                // Method:
                // 1) Get # of hyphens in original pattern
                // 2) Check that # of hyphens in the found word is equal or more
                if (hyphens_ok(p, pattern))
                  {
                  found_word(p);
                  num_words++;
                  }
                }

				      else if (p_error)
                {
			          break;
                }
              }
            }
          }
        else
          {
          found_word(p);
          num_words++;
          }
        }

      else if (anagram_error)
        {
		    gtk_label_set_text(GTK_LABEL(status_label), anagram_error);
		    g_print("%s%s\n",status_txt, anagram_error);
        return;
        }

      *t = '\n'; /* restore the newline so that it is correct for another lookup */
			t++;       /* check next word in the buffer */
			p = t;
      }
    else t++;
    }

  if (p_error)
    {
		/* Pattern error */
		sprintf(buffer, "Pattern error: %s ", pattern_errTXT[p_error]);
		gtk_label_set_text(GTK_LABEL(status_label), buffer);
		g_print("%s%s\n",status_txt, buffer);
    }

  else if (num_words == 0) strcpy(buffer, "Found no anagrams");

  else
    {
		sprintf(buffer, "Found %d anagram", num_words);
		if (num_words > 1) strcat(buffer, "s");
    }
  gtk_label_set_text(GTK_LABEL(status_label), buffer);
  g_print("%s%s\n",status_txt, buffer);
  }

else
  {
	char buffer[128];
	strcpy(buffer, "No anagram letters!");
	gtk_label_set_text(GTK_LABEL(status_label), buffer);
	g_print("%s%s\n", status_txt, buffer);
  }

}






//==============================================================================
//                      SEARCH FOR MATCHES
//==============================================================================


// We need to be able to have hyphens in pattern if desired
// A hyphen may occur anywhere except as first and last characters

void callback_search(GtkWidget *widget, gpointer data)
{
create_result_window();
if (!got_wordbuffer()) return;

if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (anagram_radio_button)))
  {
  search_for_anagrams();
  }
else
  {
	char *p;  /* position in wordbuffer */
	char *t;  /* terminal zero of current word being checked */
	char *pattern = (char *)gtk_entry_get_text(GTK_ENTRY(pattern_entry));
	gulong p_error = 0;
	int num_words = 0;
	
//	g_print("callback_search: got %s\n", pattern);
		
	
	/*
	1: Load all the words into memory
	2: Search through for all the matches an put the into a list
	3: clean the memory
	*/
	
	if (pattern && *pattern)
		{
		char buffer[strlen(pattern) + 128];

		sprintf(buffer, "Words matching pattern:-\n           \"%s\"\n", pattern);
    g_print("%s", buffer);
		append_string(buffer, 0);


	  p = t = wordbuffer;
		while (t < wordbuffer_end)
		  {
		  if (*t == '\n')
        {
				*t = '\0';

				if (match(  p,            /* word           */
								pattern,          /* pattern        */
								!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)), /*FALSE = case sensitive */
								NULL,         /* eos            */
								&p_error) )   /* error number   */
          {
					/* FOUND A WORD */
          found_word(p);
          num_words++;

          }
				else if (p_error)
          {
			    break;
          }
        

        else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(include_hyphens_check_button)))
          {
          // Copy the pattern and the word, remove hyphens from both copies then check
          char word_no_hyphen[strlen(p) + 1];
          char pattern_no_hyphen[strlen(pattern) + 1];

          char *w = word_no_hyphen;
          char *t = p;

          for (;;)   /* get the word with no hyphens */
            {
            if (*t == '-') t++;
            else
              {
              *w++ = *t;
              if (*t == 0) break;
              else t++;
              }
            }

          w = pattern_no_hyphen;
          t = pattern;

          for (;;)   /* get the pattern with no hyphens */
            {
            if (*t == '-') t++;
            else
              {
              *w++ = *t;
              if (*t == 0) break;
              else t++;
              }
            }

          if (match(  word_no_hyphen,       /* word           */
								pattern_no_hyphen,          /* pattern        */
								!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(case_sensitive_check_button)), /*FALSE = case sensitive */
								NULL,         /* eos            */
								&p_error) )   /* error number   */
              {
					    /* FOUND A WORD */
              // Re-check the hyphens here
              // Method:
              // 1) Get # of hyphens in original pattern
              // 2) Check that # of hyphens in the found word is equal or more
              if (hyphens_ok(p, pattern))
                {
                found_word(p);
                num_words++;
                }
              }

				  else if (p_error)
            {
			      break;
            }
          }





        *t = '\n'; /* restore the newline so that it is correct for another lookup */
				t++;       /* check next word in the buffer */
				p = t;
        }
			else t++;
      }
				
		if (p_error)
      {
			/* Pattern error */
			char buffer[strlen(pattern_errTXT[p_error]) + 64];
			sprintf(buffer, "Pattern error: %s", pattern_errTXT[p_error]);
			gtk_label_set_text(GTK_LABEL(status_label), buffer);
			g_print("%s%s\n",status_txt, buffer);
      }
		else
      {
			char buffer[128];
				
			if (num_words == 0) strcpy(buffer, "Found no words");
			else
				{
				sprintf(buffer, "Found %d match", num_words);
				if (num_words > 1) strcat(buffer, "es");
				}
			gtk_label_set_text(GTK_LABEL(status_label), buffer);
			g_print("%s%s\n",status_txt, buffer);
      }					
    }
	else
		{
		char buffer[128];
		strcpy(buffer, "Pattern to find is empty!");
		gtk_label_set_text(GTK_LABEL(status_label), buffer);
		g_print("%s%s\n", status_txt, buffer);
		}
  }
}


				
static void callback_with_search_check_button_clicked(GtkWidget *widget, gpointer data)
{
//printf("callback_with_search_check_button_clicked:\n");
if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (with_search_check_button)))
  {
  gtk_widget_show(pattern_label);
  gtk_widget_show(pattern_entry);
  }
else
  {
  gtk_widget_hide(pattern_label);
  gtk_widget_hide(pattern_entry);
  }
}


// ==============================
// Change anagram/search mode
// ==============================
void callback_radio_clicked(GtkWidget *widget, gpointer data)
{
g_print("callback_radio_clicked\n");


if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (widget)))
	{
	if (widget==match_radio_button)
		{
		g_print("match active\n");
				
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anagram_radio_button), FALSE);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_radio_button), TRUE);

    gtk_widget_show(pattern_label);
    gtk_widget_show(pattern_entry);
    gtk_widget_hide(anagram_label);
    gtk_widget_hide(anagram_entry);
    gtk_widget_hide(with_search_check_button);
    }
	else
		{
    g_print("anagram active\n");
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_radio_button), FALSE);
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anagram_radio_button), TRUE);

		gtk_widget_show(with_search_check_button);
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON (with_search_check_button)))
      {
			gtk_widget_show(with_search_check_button);

      gtk_widget_show(pattern_label);
      gtk_widget_show(pattern_entry);
					
    	gtk_widget_show(anagram_label);
    	gtk_widget_show(anagram_entry); 
      }
		else
			{
      gtk_widget_hide(pattern_label);
      gtk_widget_hide(pattern_entry);
	
     	gtk_widget_show(anagram_label);
    	gtk_widget_show(anagram_entry);
      }
    }
	}
}

void menuitem_response(GtkWidget *widget, gpointer gdata)
{
g_print("called up a menu\n");
}


static gint callback_menu(GtkWidget *widget, GdkEventButton *event)
{
if( (event->type == GDK_BUTTON_PRESS) && (event->button == 3) )
  {
  g_print("callback_menu\n");


  gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
                        event->button, event->time);



  return TRUE;
  }
else if (widget==close_button)
  {
  callback_quit(widget, NULL);
  return TRUE;
  }
else if (widget==preferences_button)
  {
  callback_preferences(widget, NULL);
  return TRUE;
  }
else if (widget==search_button)
  {
  callback_search(widget, NULL);
  return TRUE;
  }
return FALSE;
}

void create_menu(void)
{
int i;
    
menu = gtk_menu_new ();

for (i = 0; i < 4; i++)
  {

            /* Create a new menu-item with action name... */
  menu_item = gtk_menu_item_new_with_label (action_txt[i]);

            /* ...and add it to the menu. */
	gtk_menu_shell_append(GTK_MENU_SHELL(menu),  menu_item);
		
            /* do the action assciated with this action name */
  g_signal_connect(G_OBJECT (menu_item), "activate",
                G_CALLBACK(action[i]), NULL);


      
            /* Show the widget */
  gtk_widget_show (menu_item);
  }    
}


/**************************
 * Check for "Return" key *
 **************************/
gboolean key_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{

//cout << "GMan::key_event - gdk_keyval_name= " << gdk_keyval_name(event->keyval) << endl;

if ( strcmp(gdk_keyval_name(event->keyval), "Return" ) == 0)
	{
	if (widget == main_window)
		{
		callback_search(NULL, NULL);
		return(TRUE);  // We've done what we have to do
		}
	}

return FALSE;
}


int main(int argc, char *argv[])
{
GtkWidget *wid;
GtkWidget *box, *vbox, *hbox;
GtkWidget  *grid; //*frame_grid;
	
g_print("Hello world\n");

sleep(1);
fflush(stdin);

	
/* --- Initialize GTK --- */
gtk_init (&argc, &argv);

/* Load paths etc. */
load_prefs();

/* Load the dictionary */
wordbuffer = load_file(words_pn_txt);
if (wordbuffer)
  {
	/* address of final zero byte */
	wordbuffer_end = &wordbuffer[strlen(wordbuffer)];
  }


/* --- Create the top level window --- */
main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

gtk_container_set_border_width(GTK_CONTAINER(main_window), 0);

gtk_window_set_default_size(GTK_WINDOW(main_window), main_window_width, main_window_height); /* w, h */
gtk_window_resize(GTK_WINDOW(main_window), main_window_width, main_window_height); /* w, h */

gtk_window_set_title(GTK_WINDOW(main_window), RH_PROJECT_NAME"-"RH_PROJECT_VERSION);
g_signal_connect(G_OBJECT(main_window), "delete_event",
		      G_CALLBACK(callback_quit), NULL);

gtk_widget_add_events(main_window, GDK_BUTTON_PRESS_MASK);
g_signal_connect(G_OBJECT(main_window), "button_press_event", G_CALLBACK(callback_menu), NULL);

g_signal_connect(G_OBJECT(main_window), "key-release-event", G_CALLBACK(key_event), NULL);
mw_pixbuf = gtk_widget_render_icon_pixbuf(GTK_WIDGET(main_window),	
				    			GTK_STOCK_INDEX,
				    			GTK_ICON_SIZE_DIALOG);
	
if (mw_pixbuf) gtk_window_set_icon( GTK_WINDOW(main_window), mw_pixbuf);
			  
rw_pixbuf	= gtk_widget_render_icon_pixbuf(GTK_WIDGET(main_window),
				    				GTK_STOCK_DIALOG_INFO,
				    				GTK_ICON_SIZE_DIALOG);
	                                         	
pw_pixbuf = gtk_widget_render_icon_pixbuf(GTK_WIDGET(main_window),
 				    				GTK_STOCK_PREFERENCES,
				    				GTK_ICON_SIZE_DIALOG);
gtk_widget_show(main_window);	
create_menu();

vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
gtk_box_set_homogeneous(GTK_BOX(vbox), FALSE);
gtk_container_add(GTK_CONTAINER(main_window), vbox);	
gtk_widget_show(vbox);    

	
grid = gtk_grid_new();
//gtk_grid_set_row_homogeneous(GTK_GRID(grid), TRUE);
	
gtk_container_set_border_width(GTK_CONTAINER(grid), 10);
gtk_box_pack_start(GTK_BOX(vbox), grid, TRUE, TRUE, 0);	
gtk_widget_show(grid);    




hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
gtk_box_set_homogeneous(GTK_BOX(hbox), TRUE);
gtk_widget_show(hbox);	 
gtk_grid_attach (GTK_GRID(grid),	hbox, 135, 10, 255, 30); // L, T, W, H

	
	
// ==============
// SEARCH PATTERN
// ==============
pattern_label = gtk_label_new("Search Pattern:");
gtk_widget_show(pattern_label);
	
gtk_grid_attach (GTK_GRID(grid),	pattern_label, 10, 10, 120, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	pattern_label, 0, 0, 1, 1); // L, T, W, H

pattern_entry = gtk_entry_new();
g_signal_connect(G_OBJECT(pattern_entry), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_show(pattern_entry);
gtk_widget_set_tooltip_text(GTK_WIDGET(pattern_entry),
      "Search Pattern\n"
      "Enter the pattern you want to find.\n"
      " Use:\n"
      " *\n"
      "  for any string of characters\n"
      " ?\n"
      "  for any single character\n"
      " [xyz]\n"
      "  to search for either single character x or y or z\n"
      " [a(de)(ghi)]\n"
      "  to search for either character a or string de or string ghi");

gtk_widget_set_can_focus(GTK_WIDGET(pattern_entry), TRUE);
gtk_widget_grab_focus(GTK_WIDGET(pattern_entry));
//gtk_grid_attach (GTK_GRID(grid),	pattern_entry, 135, 10, 255, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	pattern_entry, 1, 0, 2, 1); // L, T, W, H
gtk_container_add(GTK_CONTAINER(hbox), pattern_entry);	

GtkRequisition natural_size;
gtk_widget_get_preferred_size(GTK_WIDGET(pattern_entry), NULL, &natural_size);
gtk_widget_set_size_request(GTK_WIDGET(hbox), natural_size.width, natural_size.height);
	

hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
gtk_box_set_homogeneous(GTK_BOX(hbox), TRUE);
gtk_widget_show(hbox);	 
gtk_widget_set_size_request(GTK_WIDGET(hbox), natural_size.width, natural_size.height);
gtk_grid_attach (GTK_GRID(grid),	hbox, 135, 40, 255, 30); // L, T, W, H
                                                         
	
// ===============
// ANAGRAM LETTERS
// ===============
anagram_label =  gtk_label_new("Anagram Letters:");
gtk_widget_hide(anagram_label);
//gtk_widget_show(anagram_label);

	
gtk_grid_attach (GTK_GRID(grid),	anagram_label, 10, 40, 120, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	anagram_label, 0, 1, 1, 1); // L, T, W, H

anagram_entry = gtk_entry_new();
gtk_widget_hide(anagram_entry);
//gtk_widget_show(anagram_entry);
                
g_signal_connect(G_OBJECT(anagram_entry), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);	
gtk_widget_set_tooltip_text(GTK_WIDGET(anagram_entry),
      "Anagram Letters\n"
      "Enter the letters to find anagrams thereof.\n"
      "Use ? for each unknown character.\n"
      "Use any printable character except * [ ] ( )");
//gtk_grid_attach (GTK_GRID(grid),	anagram_entry, 135, 40, 255, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	anagram_entry, 1, 1, 2, 1); // L, T, W, H
gtk_container_add(GTK_CONTAINER(hbox), anagram_entry);	





match_radio_button = gtk_radio_button_new_with_label(NULL, "Search for matches");
gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (match_radio_button), TRUE);
group = gtk_radio_button_get_group(GTK_RADIO_BUTTON (match_radio_button));
g_signal_connect(G_OBJECT(match_radio_button), "button_press_event",
                 G_CALLBACK(callback_radio_clicked), NULL);
gtk_widget_show(match_radio_button);
gtk_grid_attach (GTK_GRID(grid),	match_radio_button, 10, 80, 190, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	match_radio_button, 0, 2, 1, 1); // L, T, W, H

	

gtk_widget_set_tooltip_text(GTK_WIDGET(match_radio_button),
      "Search for matches\n"
      "Select this to search for all matches\n"
      "in the dictionary that fit the chosen pattern.");


anagram_radio_button = gtk_radio_button_new_with_label(group, "Search for anagrams");
gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (anagram_radio_button), FALSE);

gtk_widget_show(anagram_radio_button);	
g_signal_connect(G_OBJECT(anagram_radio_button), "button_press_event",
                 G_CALLBACK(callback_radio_clicked), NULL);


gtk_widget_set_tooltip_text(GTK_WIDGET(anagram_radio_button),
      "Search for anagrams\n"
      "Select this to search for all anagrams\n"
      "in the dictionary of the chosen anagram letters.");

	
gtk_grid_attach (GTK_GRID(grid),	anagram_radio_button, 200, 80, 200, 30); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	anagram_radio_button, 3, 2, 1, 1); // L, T, W, H



wid = gtk_frame_new(" Options ");
gtk_widget_show(wid);

gtk_frame_set_label_align(GTK_FRAME(wid), 0.5, 0.5);
	
gtk_grid_attach (GTK_GRID(grid),	wid, 20, 120, 380, 80); // L, T, W, H
//gtk_grid_attach (GTK_GRID(grid),	wid, 0, 3, 4, 5); // L, T, W, H



	
//frame_grid = gtk_grid_new();
//gtk_container_set_border_width(GTK_CONTAINER(frame_grid), 10);
//gtk_grid_set_column_homogeneous(GTK_GRID(frame_grid), TRUE);
//gtk_grid_set_row_homogeneous(GTK_GRID(frame_grid), TRUE);

//gtk_widget_show(frame_grid);
//gtk_container_add(GTK_CONTAINER(wid), frame_grid);

box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
gtk_container_add(GTK_CONTAINER(wid), box);	
gtk_widget_show(box);    


hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
gtk_box_pack_start(GTK_BOX(box), hbox, TRUE, TRUE, 0);	
gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);
gtk_widget_show(hbox);	

 
synonyms_check_button = gtk_check_button_new_with_label("Append synonyms");
gtk_widget_show(synonyms_check_button);
g_signal_connect(G_OBJECT(synonyms_check_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(synonyms_check_button),
      "Append synonyms\n"
      "Select this if you want words with similar meanings");

	
//gtk_grid_attach (GTK_GRID(frame_grid),	synonyms_check_button, 0, 0, 190, 1); // L, T, W, H	
//gtk_grid_attach (GTK_GRID(frame_grid),	synonyms_check_button, 0, 0, 20, 1); // L, T, W, H
gtk_box_pack_start(GTK_BOX(hbox), synonyms_check_button, TRUE, TRUE, 0);	
 

with_search_check_button = gtk_check_button_new_with_label("With Search Pattern  ");
g_signal_connect(G_OBJECT(with_search_check_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);

gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (with_search_check_button), FALSE);
g_signal_connect(G_OBJECT (with_search_check_button), "clicked",
            G_CALLBACK(callback_with_search_check_button_clicked), NULL);
gtk_widget_hide(with_search_check_button);


gtk_widget_set_tooltip_text(GTK_WIDGET(with_search_check_button),
      "With Search Pattern\n"
      "Select this to modify the search for anagrams\n"
      "to obtain only those which fit the selected pattern.");
 
	
//gtk_grid_attach (GTK_GRID(frame_grid),	with_search_check_button, 190, 0, 190, 1); // L, T, W, H
//gtk_grid_attach (GTK_GRID(frame_grid),	with_search_check_button, 2, 0, 20, 1); // L, T, W, H
//gtk_box_pack_end(GTK_BOX(hbox), with_search_check_button, TRUE, TRUE, 0);	
gtk_container_add(GTK_CONTAINER(hbox), with_search_check_button);
  

 
 

hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
gtk_box_pack_start(GTK_BOX(box), hbox, TRUE, TRUE, 0);	
gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);
gtk_widget_show(hbox);	

case_sensitive_check_button = gtk_check_button_new_with_label("Case sensitive");
gtk_widget_show(case_sensitive_check_button);
g_signal_connect(G_OBJECT(case_sensitive_check_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(case_sensitive_check_button),
      "Case sensitive\n"
      "Select this to use a case sensitive search");
                            
	
//gtk_grid_attach (GTK_GRID(frame_grid),	case_sensitive_check_button, 0, 1, 190, 1); // L, T, W, H
//gtk_grid_attach (GTK_GRID(frame_grid),	case_sensitive_check_button, 0, 1, 2, 1); // L, T, W, H
gtk_box_pack_start(GTK_BOX(hbox), case_sensitive_check_button, TRUE, TRUE, 0);	
  


include_hyphens_check_button = gtk_check_button_new_with_label("Auto-include hyphens");
gtk_widget_show(include_hyphens_check_button);
g_signal_connect(G_OBJECT(include_hyphens_check_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(include_hyphens_check_button),
      "Auto-include hyphens\n"
      "Select this to include hyphenated words\n"
      "although the pattern or anagram letters may not have hyphens.");

	
//gtk_grid_attach (GTK_GRID(frame_grid),	include_hyphens_check_button, 190, 1, 190, 1); // L, T, W, H
//gtk_grid_attach (GTK_GRID(frame_grid),	include_hyphens_check_button, 2, 1, 2, 1); // L, T, W, H
//gtk_box_pack_end(GTK_BOX(hbox), include_hyphens_check_button, TRUE, TRUE, 0);	
  
gtk_container_add(GTK_CONTAINER(hbox), include_hyphens_check_button);


hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
gtk_container_set_border_width(GTK_CONTAINER(hbox), 10);
gtk_container_add(GTK_CONTAINER(vbox), hbox);
gtk_widget_show(hbox);	


// ======
// SEARCH
// ======
search_button = gtk_button_new_with_label(action_txt[0]);
gtk_widget_show(search_button);
g_signal_connect(G_OBJECT(search_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(search_button),
      "SEARCH\n"
      "Search the dictionary for the selected pattern,\n"
      "anagrams or patterned anagrams.");
gtk_box_pack_start(GTK_BOX(hbox), search_button, TRUE, TRUE, 0);


// ===========
// PREFERENCES
// ===========
preferences_button = gtk_button_new_with_label(action_txt[1]);
gtk_widget_show(preferences_button);  
g_signal_connect(G_OBJECT(preferences_button), "button_press_event",
                   G_CALLBACK (callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(preferences_button),
      "Preferences\n"
      "Choose a dictionary and/or thesaurus\n"
      "and select window sizes");  
gtk_box_pack_start(GTK_BOX(hbox), preferences_button, TRUE, TRUE, 0);


  
// ======
// CLOSE
// ======
close_button = gtk_button_new_with_label(action_txt[3]);
gtk_widget_show(close_button);
g_signal_connect(G_OBJECT(close_button), "button_press_event",
                   G_CALLBACK(callback_menu), NULL);
gtk_widget_set_tooltip_text(GTK_WIDGET(close_button),
      "Close\n"
      "Terminate the program");
gtk_box_pack_start(GTK_BOX(hbox), close_button, TRUE, TRUE, 0);

  
gtk_main();
	
g_print("Goodbye world\n");

return (0);
}
