/****************************************************************************
*RevHead_Start                                                              *
*                                                                           *
* Project Details                                                           *
* ---------------                                                           *
*Project Name:    wordmatch                                                 *
*Project Version: 1.0                                                       *
*Copyright Owner: David M. Balean <linprogs@balean.net>                     *
*Copyright Notice:                                                          *
* This program is free software; you can redistribute it and/or modify it   *
* under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License or,         *
* (at your option) any later version.                                       *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
* See the GNU General Public License for more details.                      *
* You should have received a copy of the GNU General Public License         *
* along with this program. If not, see <http://www.gnu.org/licenses/>.      *
*                                                                           *
*Project Date:    Thursday 25th April 2013                                  *
*                                                                           *
*Start DateStamp: 17-Jun-2011 12.37.31 -36000 0                             *
*                                                                           *
*Contact:                                                                   *
* email: linprogs@balean.net                                                *
* snail: D.M. Balean, 44 Wyong Road, Killarney Vale, N.S.W. 2261, AUSTRALIA *
* URL:   http://balean.net/linprogs/index.html                              *
*                                                                           *
*Project Summary:                                                           *
* Match words using wildcards with optional synonyms                        *
* or provide anagrams using wildcards with optional synonyms                *
*                                                                           *
*****************************************************************************
*                                                                           *
*This File:       match.c                                                   *
*Author:          David M. Balean <linprogs@balean.net>                     *
*DateStamp:       18-Jun-2011 21.49.35 -36000 0                             *
*                                                                           *
*File Summary:                                                              *
* Functions for wildcard matching                                           *
*                                                                           *
*RevHead_End                                                                *
****************************************************************************/

#include <ctype.h>
#include <string.h>
#include "match.h"


static gboolean i_flag;

char *pattern_errTXT[] =
{
"OK",
"Missing character after \"\\\"",
"\"[\" without a later matching \"]\"",
"\"]\" without a previous matching \"[\"",
"Incomplete range of alternatives \"-\"",
"\"(\" without a later matching \")\"",
"\")\" without a previous matching \"(\"",

"Extraneous \"(\"",
"Extraneous \")\"",
"Extraneous \"]\"",

"Internal error",
};


gboolean do_match(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error);
gboolean sbracket(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error);
gboolean to_end_bracket(char **p_ptr, char end_bracket, long *brackets, gulong *p_error);


/**
   Check for found (i.e. both source and pattern at the end)
   Return TRUE if found
**/

gboolean matched(char *p, char *s, char *end)
{

if (*p == 0)  // finished the pattern...
   {
   if (end) return(TRUE);          // matched, source continues...
   else if (*s == 0) return(TRUE); // finished the source as well so complete match
   }
return(FALSE);
}

// Return TRUE if pattern actual character matches source actual character
gboolean match_a_char(char pc, char sc)
{
if (i_flag)
   {
   if (toupper(pc) == toupper(sc)) return(TRUE);
   }
else if (pc == sc) return(TRUE);
return(FALSE);
}

/**
   Check for end (i.e. either source or pattern at end)
   Return TRUE if at end
**/

gboolean at_end(char *p, char *s, char *end)
{
if (*p == 0) return(TRUE);
if ((end) && (s >= end)) return(TRUE);
else if (*s == 0) return(TRUE);

return(FALSE);
}

/**
   Return the number of normal characters matched
   There must be at least one normal character available
**/
// ====================================================
// CASE CHECK needed here as well - use match_a_char
// ====================================================
gboolean find_char(char **p_ptr, char **s_ptr)
{
char *p = *p_ptr;
char *s = *s_ptr;

if ( (*p) && notwildcard(*p) && (*s) && (match_a_char(*p, *s)) )
   {
   (*p_ptr)++;
   (*s_ptr)++;
   return(TRUE);
   }
return(FALSE);
}

/**
   Check literal char
**/
// ====================================================
// NOTE:- Probably put case insensitive test here also, match_a_char().
//        Case only matters with literal characters
// ====================================================
gboolean literal_char(char *p, char *s, char *end, gulong *p_error)
{
p++;
if (at_end(p, s, end))
   {
   if (*p == 0) (*p_error) = NO_CHAR_ERR;   // we are expecting a character
   return(FALSE);
   }
return(match_a_char(*p, *s));
}


/**
   Force match with zero or any number of chars
   Should come here with the character AFTER the  asterisk in p
**/

gboolean any_chars(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error)
{
  char *p = *p_ptr;
  char *s = *s_ptr;

  while (*(p) == '*') p++;
  *p_ptr = p;

  if (*p)
    {
      while ( *s )
	{
	  if (do_match(&s, &p, end, sbrackets, cbrackets, p_error))
	    {
	      if (*p == ')') // Returned from do_match because in C bracket
		{
		  p++;
		  if (*cbrackets) (*cbrackets)--;

		  else
		    { // ERROR
		      (*p_error) = C_BRACKET_ERR;
		      return(FALSE);
		    }

		  if (*sbrackets)
		    {
		      to_end_bracket(&p, ']', sbrackets, p_error); // Go to next non-paired ']'
		      continue;
		    }
		}

	      return(TRUE);
	    }
	  if (*p_error) return(FALSE);;
	  s++;
	} // END OF WHILE
      return(FALSE);
    }

  (*p_ptr) = p;
  (*s_ptr) = (end) ? end : s + strlen(s);

  return(TRUE);  // NOTE: it matches the rest of the source string
}



/**
   Move pointer to character after matching end square/C/brace bracket
**/

gboolean to_end_bracket(char **p_ptr, char end_bracket, long *brackets, gulong *p_error)
{
char *p = *p_ptr;
long numbrackets = 1;
char start_bracket;
int error;

switch(end_bracket)
   {
   case ']':
   start_bracket = '[';
   error = S_BRACKET_ERR;
   break;

   case ')':
   start_bracket = '(';
   error = C_BRACKET_ERR;
   break;

	/**
      case '}':
	start_bracket = '{';
	error = BRACE_ERR;
	break;
	**/

   default:
   (*p_error) = INTERNAL_ERR;
   return(FALSE);
   }

    
// This character matches so go to just beyond end_bracket
while (numbrackets)
   {
   UNLESS (*p)             // ERROR - Make sure there is a pattern character
      {
      (*p_error) = error;
      return(FALSE);
      }

   if (*p == '\\')
      {
      p++;                     // Skip the '\'
      UNLESS (*p)     // ERROR
         {
         (*p_error) = error;
         return(FALSE);
         }
      }
   else if (*p == start_bracket) (numbrackets)++;
   else if (*p == end_bracket) (numbrackets)--;
   p++;                             // Skip this pattern character
   }

(*brackets)--; // either sbrackets or cbrackets

*p_ptr = p;
return(TRUE);;
}


/****
     Check out an alternative sub_string
     If found, it must check to the very end from here aaand return TRUE if found
     OTHERWISE it MUST go to end bracket
****/

gboolean cbracket(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error)
{
  char *p = *p_ptr;
  char *s = *s_ptr;
  long cb_save = ++(*cbrackets);
  long sb_save = (*sbrackets);

  p++; // Skip the C bracket

  UNTIL ((*p) == ')')
    {

      UNLESS(*p)
	{
	  (*p_error) = C_BRACKET_ERR;
	  return(FALSE);
	}
      else if (at_end(p, s, end))
	{
	  return(FALSE); // try next alternative if there is one (e.g. more sbrackets)
	}

      switch(*p)
	{
	case '?':
	  p++;
	  s++;
	  break;
	   
	case '\\':
	  UNLESS (literal_char(p, s, end, p_error)) return(FALSE);
	  p += 2;
	  s++;
	  break;

	case '*':
	  p++;
	  if (any_chars( &s, &p, end, sbrackets, cbrackets, p_error))
	    {
 	      (*p_ptr) = p + strlen(p);
	      (*s_ptr) =  s + strlen(s);
	      return(TRUE);
	    }
	  if (*p_error) return(FALSE);

	  return(FALSE);

	  
	case '[':
	  // NOTE: sbracket() should check to the very end
	  UNLESS (sbracket( &s, &p, end, sbrackets, cbrackets, p_error))
	    {
	      return(FALSE);
	    }

          (*p_ptr) = p;
          (*s_ptr) = s;
          return(TRUE);
	     
	  
	case '(':
	  (*p_error) = BAD_C1BRACKET_ERR;
	  return(FALSE);
	  

	  /*** Shouldn't be possible!
	case ')':
	  (*cbrackets)--;
	  p++;
	  break;
	  ***/


	case ']':
	  (*p_error) = C_BRACKET_ERR;
	  return(FALSE);

	default: // any other character
	  UNLESS (find_char(&p, &s)) return(FALSE); // quit if not a match
	  break;
	  
	}

    } // END OF UNTIL ((*p) == ')')

  
  p++;                // pass over the C bracket
  (*cbrackets)--;

  UNLESS (to_end_bracket(&p, ']', sbrackets, p_error)) return(FALSE);
  if (do_match(&s, &p, end, sbrackets, cbrackets, p_error))
    {
      (*p_ptr) = p + strlen(p);
      (*s_ptr) =  s + strlen(s);
      return(TRUE);
    }
  else
    {
      (*sbrackets) = sb_save;
      (*cbrackets) = cb_save;
    }
  return(FALSE);
}

/**
   Match an alternative from within within square brackets
   Return TRUE if an alternative matches
   There appears to be a problem in that the order of
   the pattern can influence the output.  In a match is
   obtained but there is no overall match the next alternative
   should be checked and so on until there are no more choices.
   This means re-calling do_match().
**/

gboolean sbracket(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error)
{
char *p_save;           // save position in pattern string when skipping rest of pattern
char *s_save;           // save position in source string when skipping rest of pattern

char *p = *p_ptr;        // Easier to understand this way...
char *s = *s_ptr;
long sb_save = ++(*sbrackets);
long cb_save = *cbrackets;

p++;

UNTIL (*p == ']')
   {
   UNLESS (*p)
	   {
	   (*p_error) = S_BRACKET_ERR;
	   return(FALSE);
	   }
   else if (at_end(p, s, end))
	   {
	   return(FALSE); // try next alternative if there is one (e.g. more sbrackets)
	   }

   p_save = p;               // save position in pattern string when skipping rest of pattern
   s_save = s;               // save position in source string when skipping rest of pattern


   // We have both p and s

   switch(*p)
	   {
	   case '?':
      s++; // FORCE A MATCH
      to_end_bracket(&p, ']', sbrackets, p_error); // Go to next non-paired ']'
      if (*p_error) return(FALSE);

      if (do_match(&s, &p, end, sbrackets, cbrackets, p_error))
	      {
		   (*p_ptr) = p + strlen(p);
		   (*s_ptr) = (end) ? end : (s + strlen(s));
		   return(TRUE);
	      }
      if (*p_error) return(FALSE);

      // It didn't match so try next char
      s = s_save;
      p_save++;
      *sbrackets = sb_save;
      *cbrackets = cb_save;
      break;
	   
	   case '\\':
      p_save++;

      if (literal_char(p, s, end, p_error))
	      {
		   p += 2;
		   to_end_bracket(&p, ']', sbrackets, p_error);
		   if (*p_error) return(FALSE);
		   s++;
		   if (do_match(&s, &p, end, sbrackets, cbrackets, p_error))
		      {
		      (*p_ptr) = p + strlen(p);
		      (*s_ptr) =   (end) ? end : (s + strlen(s));
		      return(TRUE);
		      }
         }
      if (*p_error) return(FALSE);

      s = s_save;
      p_save++;
      *sbrackets = sb_save;
      *cbrackets = cb_save;
      break;
	    
	   case '*':
      to_end_bracket(&p, ']', sbrackets, p_error);
      if (*p_error) return(FALSE);

      if (any_chars(&s, &p, end, sbrackets, cbrackets, p_error))
	      {
		   (*p_ptr) = p + strlen(p);
		   (*s_ptr) = (end) ? end : (s + strlen(s));
		   return(TRUE);
	      }
      if (*p_error) return(FALSE);
      p_save++;
      s = s_save;
      *sbrackets = sb_save;
      *cbrackets = cb_save;
      break;

	   case '[':
      if (sbracket( &s, &p, end, sbrackets, cbrackets, p_error))
	      {
		   to_end_bracket(&p, ']', sbrackets, p_error); // Go to next non-paired ']'
         (*p_ptr) = p;
		   (*s_ptr) = s;
         return(TRUE);  // NOTE: this doesn't zero sbrackets
         }


      (*sbrackets) = sb_save;
      (*cbrackets) = cb_save;

      // We need to skip this sbracket as it failed
      to_end_bracket(&p, ']', sbrackets, p_error);

      if (*p_error) return(FALSE);
	    
      sb_save = (*sbrackets) - 1;
      cb_save = (*cbrackets);
      return(FALSE);    // I THINK THAT IF AT END WE SHOULD RETURN!

	   case '(':
      if (cbracket(&s, &p, end, sbrackets, cbrackets, p_error)) // it matched
	      {
		   (*p_ptr) = p;
		   (*s_ptr) = s;
		   return(TRUE);
	      }
 
      if (*p_error) return(FALSE);

      p++; // Skip the '(', *cbrackets already incremented
	    
      to_end_bracket(&p, ')', cbrackets, p_error);
      if (*p_error) return(FALSE);

      p_save = p;
      s = s_save;
      sb_save = *sbrackets;
      break;

	   case ')':
      (*p_error) = PREV_CBRACKET_ERR;
      return(FALSE);

      /*** Can't be this!!
	   case ']':
      (*sbrackets)--;
      p++;
      p_save = p;
      break;
      ***/


	   default:
         {
	      gboolean rangefound = FALSE;
	    
	      // Check for a range first
	      if (*(p+1) && (*(p+1)=='-'))
	         {
		      char c, start_c, end_c;
		
		      UNLESS (*(p+2))
		         {
		         (*p_error) = INCOMPLETE_RANGE_ERR;
		         return(FALSE);
		         }
		      if ( (*p) < (*(p+2)) )
		         {
		         start_c = *p;
		         end_c = (*(p+2));
		         }
		      else
		         {
		         start_c = (*(p+2));
		         end_c = *p;
		         }

		      for (c=start_c; c<=end_c; c++)
		         {
//               if (c == *s)  // Do we worry about CASE here?
               if(match_a_char(c, *s))
                  {
			         p_save += 2;
			         p += 2;
			         rangefound = TRUE;
			         break;
                  }
		         }
	         }

         // ======================
         // CASE CHECK NEEDED HERE
         // ======================
	      //if (rangefound || ((*p == *s) && (*(p+1) != '-')) )   // -----------CASE?
	      if (rangefound || ((match_a_char(*p, *s)) && (*(p+1) != '-')) )
            {
		      s++;
		      to_end_bracket(&p, ']', sbrackets, p_error); // Go to next non-paired ']'
		      if (*p_error) return(FALSE);

		      if (*sbrackets) // If we were called recursively from sbrackets
		         {
		         (*p_ptr) = p;
		         (*s_ptr) = s;
		         return(TRUE);
		         }

		      if (do_match(&s, &p, end, sbrackets, cbrackets, p_error))
		         {
               if (*p == ')')  // I think it only gets here from any_chars() via do_match()
		            {
			         (*p_ptr) = p;
			         (*s_ptr) = s;
			         return(TRUE);
		            }
 
		         (*p_ptr) = p + strlen(p);
		         (*s_ptr) =   (end) ? end : (s + strlen(s));
		         return(TRUE);
		         }

		      if (*p_error) return(FALSE);;

		      // It didn't match so try next char

	         }

	      s = s_save;
	      p_save++;
	      *sbrackets = sb_save;
	      *cbrackets = cb_save;
	      break;

	      }

      }   // END OF switch (*p)

   p = p_save;             // recover position in pattern ready for next iteration

   }   // END OF UNTIL (*p == ']')

return(FALSE);
} 

/***************************************************************
 * Return TRUE if both zero terminated strings match.          *
 ***************************************************************/

gboolean do_match(char **s_ptr, char **p_ptr, char *end, long *sbrackets, long *cbrackets, gulong *p_error)
{
char *p = *p_ptr;
char *s = *s_ptr;

FOREVER
   {
   if (at_end(p, s, end))
      {
	   if (matched(p, s, end))
         {
	      *s_ptr = s;
	      *p_ptr = p;
	      return(TRUE);
	      }
	   UNLESS (*p) break;
	   }

    /*
      TO GET HERE WE HAVE TO HAVE *s
    */

    switch (*p)
      {
      case '\0':
	   if (matched(p, s, end))
	      {
	      *s_ptr = s;
	      *p_ptr = p;
	      return(TRUE);
	   }
	   break;


	   // MATCH ANY 1 CHARACTER
      case '?':
      p++;
	   s++;
	   break;
	
	   // MATCH THE FOLLOWING LITERAL CHARACTER (useful for embedded wildcards)
      case '\\':
	   UNLESS (literal_char(p, s, end, p_error)) return(FALSE);
	   p+=2;
	   s++;
	   break;

	   // MATCH ZERO OR ANY NUMBER OF CHARACTERS
      case  '*':
	   p++;
	   return(any_chars(&s, &p, end, sbrackets, cbrackets, p_error));

	   // MATCH ALTERNATIVES
      case '[':
	   UNLESS (sbracket( &s, &p, end, sbrackets, cbrackets, p_error)) return(FALSE);
	   break;

      case '(':
	   (*p_error) = BAD_C1BRACKET_ERR;
	   break;

      case ')':
	   if (*cbrackets)  // We came here via any_chars()
	      {
	      *s_ptr = s;
	      *p_ptr = p;
	      return(TRUE);  // returns with ')' in p
	   }
	   (*p_error) = BAD_C2BRACKET_ERR;
	   break;

      case ']':
	   (*p_error) = BAD_S2BRACKET_ERR;
	   break;

      default: // any other character

	   UNLESS (find_char(&p, &s))
	      {
	      return(FALSE); // quit if not a match
	      }

      } // END OF switch (*p)


   if ( (*p_error) != 0) break;

   } // END OF FOREVER

return(FALSE);
}




gboolean sbracket_bad(char **pattern, long *sb, long *cb, gulong *p_error);

/****
     Return TRUE if there is a backslash error
****/

gboolean backslash_bad(char *p, gulong *p_error)
{
UNLESS (*(p+1))
   {
   (*p_error) = NO_CHAR_ERR;
   return(TRUE);
   }

return(FALSE);
}

/****
     Return TRUE if there is a C bracket error
****/

gboolean cbracket_bad(char **pattern, long *sb, long *cb, gulong *p_error)
{
char *p = (*pattern);

(*cb)++;
p++;

UNTIL (*p == ')')
   {
   switch (*p)
      {
      case 0:
	   (*p_error) = C_BRACKET_ERR;
	   return(TRUE);

      case '\\':
	   if (backslash_bad(p, p_error)) return(TRUE);
	   p += 2;
	   break;

      case '[':
	   if (sbracket_bad(&p, sb, cb, p_error))
         {
	      return(TRUE);
	      }
	   break;

      case ']':
	   (*p_error) = C_BRACKET_ERR;
	   return(TRUE);

      case '(':
	   (*p_error) = BAD_C1BRACKET_ERR;
	   return(TRUE);
	
      default:  // any other character
	   p++;
	   break;
      }

   } // END OF UNTIL

(*cb)--;
p++;
(*pattern) = p;
return(FALSE);
}

/****
     Return TRUE if there is a square bracket error
****/

gboolean sbracket_bad(char **pattern, long *sb, long *cb, gulong *p_error)
{
char *p = (*pattern);

(*sb)++;
p++;

UNTIL (*p == ']')
   {
   switch (*p)
      {
      case 0:
	   (*p_error) = S_BRACKET_ERR;
	   return(TRUE);

      case '\\':
	   if (backslash_bad(p, p_error)) return(TRUE);
	   p += 2;
	   break;

      case '[':
	   if (sbracket_bad(&p, sb, cb, p_error)) return(TRUE);
	   break;

      case '(':
	   if (cbracket_bad(&p, sb, cb, p_error))
	    {
	    return(TRUE);
	    }
	   break;

      case ')':
	   (*p_error) = PREV_CBRACKET_ERR;
	   return(TRUE);

      default: // any other character
	   p++;
	   break;
      }
   }

(*sb)--;
p++;
(*pattern) = p;
return(FALSE);
}


/****
     Return TRUE if there is a simple pattern error
****/

gboolean pattern_error(char *p, gulong *p_error)
{
long sb;
long cb;

sb = cb = 0;

while (*p)
   {
   switch (*p)
	   {
	   case '[':
      if (sbracket_bad(&p, &sb, &cb, p_error)) return(TRUE);
	   break;

	   case '\\':
	   if (backslash_bad(p, p_error)) return(TRUE);
	   p += 2;
	   break;

      case '(':
	   (*p_error) = BAD_C1BRACKET_ERR;
	   return(TRUE);

      case ')':
	   (*p_error) = BAD_C2BRACKET_ERR;
	   return(TRUE);

      case ']':
	   (*p_error) = BAD_S2BRACKET_ERR;
	   return(TRUE);

	   default: // any other character
	   p++;
	   break;
	   }
   }
   
(*p_error) = NO_PATTERN_ERROR;
return(FALSE); // pattern OK
}

/**********************************
The entry point
===============
s        - string start
p        - pattern
i        - if TRUE use case insensitive test
end      - end of string
p_error  - pointer to error number
**********************************/

gboolean match(char *s, char *p, gboolean i, char *end, gulong *p_error)
{
long sbrackets, cbrackets;

i_flag = i;

if (pattern_error(p, p_error))  // Currently this doesn't check range errors
   {
//      fprintf(stderr, "\n<<**- FROM pattern_error(): -**>>");
   return(FALSE);
   }

sbrackets = cbrackets = 0;

return(do_match(&s, &p, end, &sbrackets, &cbrackets, p_error));
}
