/****************************************************************************
*RevHead_Start                                                              *
*                                                                           *
* Project Details                                                           *
* ---------------                                                           *
*Project Name:    wordmatch                                                 *
*Project Version: 1.0                                                       *
*Copyright Owner: David M. Balean <linprogs@balean.net>                     *
*Copyright Notice:                                                          *
* This program is free software; you can redistribute it and/or modify it   *
* under the terms of the GNU General Public License as published by         *
* the Free Software Foundation, either version 3 of the License or,         *
* (at your option) any later version.                                       *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                      *
* See the GNU General Public License for more details.                      *
* You should have received a copy of the GNU General Public License         *
* along with this program. If not, see <http://www.gnu.org/licenses/>.      *
*                                                                           *
*Project Date:    Thursday 25th April 2013                                  *
*                                                                           *
*Start DateStamp: 17-Jun-2011 12.37.31 -36000 0                             *
*                                                                           *
*Contact:                                                                   *
* email: linprogs@balean.net                                                *
* snail: D.M. Balean, 44 Wyong Road, Killarney Vale, N.S.W. 2261, AUSTRALIA *
* URL:   http://balean.net/linprogs/index.html                              *
*                                                                           *
*Project Summary:                                                           *
* Match words using wildcards with optional synonyms                        *
* or provide anagrams using wildcards with optional synonyms                *
*                                                                           *
*****************************************************************************
*                                                                           *
*This File:       prefs.c                                                   *
*Author:          David M. Balean <linprogs@balean.net>                     *
*DateStamp:       25-Apr-2013 12.21.44 -36000 0                             *
*                                                                           *
*File Summary:                                                              *
* Obtain user preferences                                                   *
*                                                                           *
*RevHead_End                                                                *
****************************************************************************/

#include <stdlib.h>

#include "misc.h"
#include "prefs.h"

extern GtkWidget *main_window;
extern GList *child_windows;

static GtkWidget *window;

static GtkWidget *words_pn_entry             = NULL;
static GtkWidget *synonyms_command_entry     = NULL;

char *words_pn_txt                           = NULL;
char *synonyms_command_txt                   = NULL;

static GtkWidget *main_window_width_spinner     = NULL;
static GtkWidget *main_window_height_spinner    = NULL;
static GtkWidget *prefs_window_width_spinner    = NULL;
static GtkWidget *prefs_window_height_spinner   = NULL;
static GtkWidget *about_window_width_spinner    = NULL;
static GtkWidget *about_window_height_spinner   = NULL;
static GtkWidget *results_window_width_spinner  = NULL;
static GtkWidget *results_window_height_spinner = NULL;

int main_window_width                        = 0;
int main_window_height                       = 0;
int prefs_window_width                       = 0;
int prefs_window_height                      = 0;
int about_window_width                       = 0;
int about_window_height                      = 0;
int results_window_width                     = 0;
int results_window_height                    = 0;

static const int d_main_window_width         = 400;
static const int d_main_window_height        = 240;
static const int d_prefs_window_width        = 500;
static const int d_prefs_window_height       = 350;
static const int d_about_window_width        = 600;
static const int d_about_window_height       = 350;
static const int d_results_window_width      = 850;
static const int d_results_window_height     = 350;

static char *words_default_pn_txt            = "/usr/share/dict/words";
static char *synonyms_default_command_txt    = "/usr/bin/aiksaurus %s";

static char *wordmatch_prefs_pn_txt          = NULL;

char *getenv(const char *name);

void callback_preferences_cancel(GtkWidget *widget,gpointer data)
{
gtk_widget_destroy(GTK_WIDGET(window));
}

/* Write prefs file and make the values current */
void callback_preferences_ok(GtkWidget *widget, gpointer data)
{
GList *node = child_windows;
GList *nextnode;
int fd;
char buffer[8192];

char *w_txt = (char*)gtk_entry_get_text(GTK_ENTRY(words_pn_entry));
char *s_txt = (char *)gtk_entry_get_text(GTK_ENTRY(synonyms_command_entry));

main_window_width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(main_window_width_spinner));
main_window_height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(main_window_height_spinner));
prefs_window_width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs_window_width_spinner)); 
prefs_window_height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(prefs_window_height_spinner));
about_window_width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(about_window_width_spinner)); 
about_window_height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(about_window_height_spinner));
results_window_width = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(results_window_width_spinner));
results_window_height = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(results_window_height_spinner));

gtk_window_set_default_size(GTK_WINDOW(main_window), main_window_width, main_window_height); /* w, h */


while (node)
	{
	nextnode = node->next;
	gtk_window_set_default_size(GTK_WINDOW(*(GtkWindow **)(node->data)), results_window_width, results_window_height);    /* w, h */
	node = nextnode;
	}
	
sprintf(buffer,
	"# preferences file for wordmatch\n\nWORDS_PN = \"%s\"\nSYNONYMS_COMMAND = \"%s\"\n",
	w_txt, s_txt);
sprintf(&buffer[strlen(buffer)], "MAIN_W = %d\nMAIN_H = %d\nPREFS_W = %d\nPREFS_H = %d\n",
  main_window_width, main_window_height, prefs_window_width, prefs_window_height);

sprintf(&buffer[strlen(buffer)], "ABOUT_W = %d\nABOUT_H = %d\nRES_W = %d\nRES_H = %d\n",
  about_window_width, about_window_height, results_window_width, results_window_height);

fd = open(wordmatch_prefs_pn_txt,
			O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
if (fd >= 0)
	{
	if (write(fd, buffer, strlen(buffer)) == strlen(buffer))
		{
		/* ALL OK */
    sprintf(buffer, "Wrote %s", wordmatch_prefs_pn_txt);
		}
	else sprintf(buffer, "FAILED to write %s\n", wordmatch_prefs_pn_txt);
	}
else sprintf(buffer, "FAILED to open %s for write\n", wordmatch_prefs_pn_txt);

g_print("%s%s\n","STATUS: ", buffer);

if (fd >= 0) close(fd);
callback_preferences_cancel(widget, data);	
}



void callback_default_main_window_size(GtkWidget *widget, gpointer data)
{
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(main_window_width_spinner), d_main_window_width);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(main_window_height_spinner), d_main_window_height);
}

void callback_default_prefs_window_size(GtkWidget *widget, gpointer data)
{
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs_window_width_spinner), d_prefs_window_width);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs_window_height_spinner), d_prefs_window_height);
}
void callback_default_about_window_size(GtkWidget *widget, gpointer data)
{
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(about_window_width_spinner), d_about_window_width);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(about_window_height_spinner), d_about_window_height);
}
void callback_default_results_window_size(GtkWidget *widget, gpointer data)
{
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(results_window_width_spinner), d_results_window_width);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(results_window_height_spinner), d_results_window_height);
}


void callback_preferences(GtkWidget *widget, gpointer data)
{
GtkWidget *grid;
GtkWidget *wid;

window = gtk_dialog_new();

gtk_window_set_title (GTK_WINDOW (window), "Preferences");

gtk_window_set_modal(GTK_WINDOW(window), TRUE);
gtk_window_set_default_size(GTK_WINDOW(window), prefs_window_width, prefs_window_height); /* w, h */
gtk_window_resize(GTK_WINDOW(window), prefs_window_width, prefs_window_height); /* w, h */
if (pw_pixbuf) gtk_window_set_icon( GTK_WINDOW(window), pw_pixbuf);


grid = gtk_grid_new ();

gtk_container_set_border_width(GTK_CONTAINER (grid), 10);
gtk_container_add(GTK_CONTAINER(gtk_dialog_get_content_area(GTK_DIALOG(window))), grid);
wid = gtk_label_new("Words pathname:");
gtk_grid_attach (GTK_GRID(grid), wid, 10, 10, 190, 40); // L, T, W, H
words_pn_entry = gtk_entry_new();
gtk_entry_set_text(GTK_ENTRY(words_pn_entry), words_pn_txt);
gtk_widget_set_tooltip_text(GTK_WIDGET(words_pn_entry),
      "Words pathname\n"
      "Enter the full pathname of the dictionary to search.");
                            
gtk_grid_attach (GTK_GRID(grid), words_pn_entry, 210, 10, 280, 40); // L, T, W, H

wid = gtk_label_new("Synonyms command:");
gtk_grid_attach (GTK_GRID(grid), wid, 10, 55, 190, 40); // L, T, W, H
synonyms_command_entry = gtk_entry_new();
gtk_entry_set_text(GTK_ENTRY(synonyms_command_entry), synonyms_command_txt);
gtk_widget_set_tooltip_text(GTK_WIDGET(synonyms_command_entry),
      "Synonyms command\n"
      "Enter the full pathname of the synonyms program.\n"
      "Use %s for the word being sought in the thesaurus.");
gtk_grid_attach (GTK_GRID(grid), synonyms_command_entry, 210, 55, 280, 40); // L, T, W, H

 
wid = gtk_label_new("WIDTH");
gtk_misc_set_alignment(GTK_MISC (wid), .5, 1);
gtk_grid_attach (GTK_GRID(grid), wid, 210, 100, 70, 40); // L, T, W, H

wid = gtk_label_new("HEIGHT");
gtk_misc_set_alignment(GTK_MISC (wid), .5, 1);
gtk_grid_attach (GTK_GRID(grid), wid, 290, 100, 90, 40); // L, T, W, H


  
// MAIN WINDOW
//============
wid = gtk_label_new("Main Window");
gtk_misc_set_alignment(GTK_MISC (wid), 0, .5);
gtk_grid_attach (GTK_GRID(grid), wid, 10, 140, 190, 40); // L, T, W, H

main_window_width_spinner = create_spinner(main_window_width, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), main_window_width_spinner, 210, 140, 70, 40); // L, T, W, H


main_window_height_spinner =  create_spinner(main_window_height, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), main_window_height_spinner, 290, 140, 90, 40); // L, T, W, H
 
wid = gtk_button_new_with_label ("DEFAULT");
g_signal_connect (G_OBJECT (wid), "clicked",
		    G_CALLBACK ( callback_default_main_window_size ),  NULL);
gtk_grid_attach (GTK_GRID(grid), wid, 390, 140, 100, 40); // L, T, W, H

// PREFERENCES WINDOW
//===================
wid = gtk_label_new("Preferences Window");
gtk_misc_set_alignment(GTK_MISC (wid), 0, .5);
gtk_grid_attach (GTK_GRID(grid), wid, 10, 180, 190, 40); // L, T, W, H
	
prefs_window_width_spinner = create_spinner(prefs_window_width, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), prefs_window_width_spinner, 210, 180, 70, 40); // L, T, W, H
	
prefs_window_height_spinner =  create_spinner(prefs_window_height, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), prefs_window_height_spinner, 290, 180, 90, 40); // L, T, W, H
	
wid = gtk_button_new_with_label ("DEFAULT");
g_signal_connect (G_OBJECT (wid), "clicked",
		    G_CALLBACK ( callback_default_prefs_window_size ),  NULL);
gtk_grid_attach (GTK_GRID(grid), wid, 390, 180, 100, 40); // L, T, W, H

// ABOUT WINDOW
//===============
wid = gtk_label_new("About Window");
gtk_misc_set_alignment(GTK_MISC (wid), 0, .5);
gtk_grid_attach (GTK_GRID(grid), wid, 10, 220, 190, 40); // L, T, W, H

about_window_width_spinner = create_spinner(about_window_width, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), about_window_width_spinner, 210, 220, 70, 40); // L, T, W, H
	
about_window_height_spinner =  create_spinner(about_window_height, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), about_window_height_spinner, 290, 220, 90, 40); // L, T, W, H

	
wid = gtk_button_new_with_label ("DEFAULT");
g_signal_connect (G_OBJECT (wid), "clicked",
		    G_CALLBACK ( callback_default_about_window_size ),  NULL);
gtk_grid_attach (GTK_GRID(grid), wid, 390, 220, 100, 40); // L, T, W, H

    
// RESULTS WINDOW
//===============
wid = gtk_label_new("Results Window");
gtk_misc_set_alignment(GTK_MISC (wid), 0, .5);
gtk_grid_attach (GTK_GRID(grid), wid, 10, 260, 190, 40); // L, T, W, H
    
results_window_width_spinner = create_spinner(results_window_width, 0, 10000);    
gtk_grid_attach (GTK_GRID(grid), results_window_width_spinner, 210, 260, 70, 40); // L, T, W, H
    
results_window_height_spinner =  create_spinner(results_window_height, 0, 10000);
gtk_grid_attach (GTK_GRID(grid), results_window_height_spinner, 290, 260, 90, 40); // L, T, W, H
    
wid = gtk_button_new_with_label ("DEFAULT");
g_signal_connect (G_OBJECT (wid), "clicked",
		    G_CALLBACK ( callback_default_results_window_size ),  NULL);
gtk_grid_attach (GTK_GRID(grid), wid, 390, 260, 100, 40); // L, T, W, H


gtk_spin_button_set_value(GTK_SPIN_BUTTON(main_window_width_spinner), main_window_width);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(main_window_height_spinner), main_window_height);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs_window_width_spinner), prefs_window_width);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefs_window_height_spinner), prefs_window_height);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(about_window_width_spinner), about_window_width);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(about_window_height_spinner), about_window_height);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(results_window_width_spinner), results_window_width);
gtk_spin_button_set_value(GTK_SPIN_BUTTON(results_window_height_spinner), results_window_height);

wid = gtk_button_new_with_label("OK");
g_signal_connect(G_OBJECT(wid), "clicked",
		    	G_CALLBACK(callback_preferences_ok),  NULL);
gtk_box_pack_start(GTK_BOX(gtk_dialog_get_action_area(GTK_DIALOG(window))), wid, TRUE, TRUE, 0);	
gtk_widget_set_tooltip_text(GTK_WIDGET(wid),
      "OK\n"
      "Use the selected dictionary and thesaurus.");

wid = gtk_button_new_with_label("CANCEL");
gtk_widget_set_tooltip_text(GTK_WIDGET(wid),
      "CANCEL\n"
      "Forget these selections.");
g_signal_connect(G_OBJECT(wid), "clicked",
		    	G_CALLBACK(callback_preferences_cancel), NULL);				
gtk_box_pack_start(GTK_BOX(gtk_dialog_get_action_area(GTK_DIALOG(window))), wid, TRUE, TRUE, 0);
gtk_widget_show_all(window);
}




/* Create a string in memory, needs to be g_freed later */
char *make_string(char *string)
{
if (string)
  {
	char *s = (char *)g_malloc( strlen(string) + 1 );
	
	if (s)
	  {
	    strcpy(s, string);
	  }
	return(s);
  }

return(NULL);
}

/*
 Remove the quotes from the supplied string
 Two consecutive quotes represent a single literal quote
 Return the string in memory
*/
char *dequote_string(char *string)
{
char buffer[strlen(string) + 16];
char *s = string;
char *d = buffer;
gboolean in_quote = FALSE;
	
while (*s)
	{
	if (*s == '\"')
		{
		if (in_quote && (*(s+1) == '\"'))
			{
//			g_print("dequote_string - %c\n", *s);				
			*d++ = *s++;
			}				
		else
			{
//			g_print("dequote_string - changing in_quote\n");
			in_quote ^= 1;
			}
		}
	else
		{
//		g_print("dequote_string - %c\n", *s);				
		*d++ = *s;
		}
	s++;
	}
*d = 0;
	
//g_print("dequote_string - string= %s : buffer= <%s>\n", string, buffer);
return(make_string(buffer));
}


char *get_right(char *c, char *old_end)
{
char *end;
gboolean in_quote = FALSE;

/* Remove leading white space */
while ( (*c == ' ') || (*c == '\t') ) c++;   /* white space */

/* Skip an assignment character */
if ((*c == '=') || (*c == ':'))
	{
	c++;

	/* Skip white space after assignment character */
	while ( (*c == ' ') || (*c == '\t') ) c++;   /* white space */
    }

/* Now look for end of line starting here, or start of a comment */
end = c;

/*
 Find the end, comments are # or ; (Assumes no white space in the value and line ends with newline)
If old_end is defined, the character is the old character at the end which is replaced by the zero
*/
	
while ( (*end) && ( (in_quote && (*end != '\n')) ||
		(!in_quote && (*end != ' ') && (*end != '\t') && (*end != '#') && (*end != ';') && (*end != '\n') ) ) )
	{
//	g_print("GOT %c : in_quote=%d\n", *end, in_quote);
	if (*end == '\"')
		{
		if (in_quote && (*(end+1) == '\"')) end++;  /* literal */
		else in_quote ^= 1;
		}
	end++;
	}
  
if (old_end) *old_end = *end;  /* The character that was replaced by a zero */
*end = 0;

return(c);
}




void load_prefs(void)
{
char *homedir = getenv("HOME");

char prefsname[strlen(homedir) + 64];
char *buffer;
char *c;
char end;
gboolean mwf, mhf, pwf, phf, awf, ahf, rwf, rhf;
mwf = mhf = pwf = phf = awf = ahf =rwf = rhf = FALSE;
	
sprintf(prefsname, "%s/.wordmatch", homedir);
wordmatch_prefs_pn_txt = make_string(prefsname);
buffer = load_file(prefsname);
if (buffer)
	{
	c = buffer;
	while (*c)
		{
		while ( (*c == ' ') || (*c == '\t') ) c++;   /* white space */
    	if ( (*c == '#') || (*c == ';') || (*c == '\n') )
			{
			/* comments, newlines */
			while (*c && *c != '\n') c++;
			if (*c) c++;
			}

		else if (STRIEQN(c, "WORDS_PN", 8))
			{
	  	c += 8;
	  	c = get_right(c, &end);
			if (*c)
				{
				if (words_pn_txt) g_free(words_pn_txt);
				words_pn_txt = dequote_string(c);
				}
			c += strlen(c);
			*c = end;
			}
		else if (STRIEQN(c, "SYNONYMS_COMMAND", 16))
			{
	  	c += 16;
	  	c = get_right(c, &end);
			if (*c)
				{
				if (synonyms_command_txt) g_free(synonyms_command_txt);
				synonyms_command_txt = dequote_string(c);
				}
			c += strlen(c);
			*c = end;
			}
		else if (STRIEQN(c, "MAIN_W", 6))
      {
	  	c += 6;
	  	c = get_right(c, &end);
			if (*c)
        {
        main_window_width = atoi(c);
        mwf = TRUE;
        }        
			c += strlen(c);
			*c = end;
      }
 		else if (STRIEQN(c, "MAIN_H", 6))
      {
	  	c += 6;
	  	c = get_right(c, &end);
			if (*c)
        {
        main_window_height = atoi(c);
        mhf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }
 		else if (STRIEQN(c, "PREFS_W", 7))
      {
	  	c += 7;
	  	c = get_right(c, &end);
			if (*c)
        {
        prefs_window_width = atoi(c);
        pwf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }  
 		else if (STRIEQN(c, "PREFS_H", 7))
      {
	  	c += 7;
	  	c = get_right(c, &end);
			if (*c)
        {
        prefs_window_height = atoi(c);
        phf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }  
 		else if (STRIEQN(c, "ABOUTS_W", 7))
      {
	  	c += 7;
	  	c = get_right(c, &end);
			if (*c)
        {
        about_window_width = atoi(c);
        awf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }  
 		else if (STRIEQN(c, "ABOUT_H", 7))
      {
	  	c += 7;
	  	c = get_right(c, &end);
			if (*c)
        {
        about_window_height = atoi(c);
        ahf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }  
 		else if (STRIEQN(c, "RES_W", 5))
      {
	  	c += 5;
	  	c = get_right(c, &end);
			if (*c)
        {
        results_window_width = atoi(c);
        rwf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }
 		else if (STRIEQN(c, "RES_H", 5))
      {
	  	c += 5;
	  	c = get_right(c, &end);
			if (*c)
        {
        results_window_height = atoi(c);
        rhf = TRUE;
        }
			c += strlen(c);
			*c = end;
      }      
   
    else
			{
			/* UNRECOGNIZED, ignore to eol */
			while (*c && *c != '\n') c++;
			}
		}
	g_free(buffer);
	}

// Fix up defaults if not found
if (!words_pn_txt)
		words_pn_txt = make_string(words_default_pn_txt);
if (!synonyms_command_txt)
		synonyms_command_txt = make_string(synonyms_default_command_txt);
if (!(mwf)) main_window_width = d_main_window_width;
if (!(mhf)) main_window_height = d_main_window_height;
if (!(pwf)) prefs_window_width = d_prefs_window_width;
if (!(phf)) prefs_window_height = d_prefs_window_height;
if (!(awf)) about_window_width = d_about_window_width;
if (!(ahf)) about_window_height = d_about_window_height;   
if (!(rwf)) results_window_width = d_results_window_width;
if (!(rhf)) results_window_height = d_results_window_height;
}

